@@ -1,182 +1,13 @@
 package Config::IOD::Reader;
 
-our $DATE = '2015-01-03'; # DATE
-our $VERSION = '0.07'; # VERSION
+our $DATE = '2015-03-12'; # DATE
+our $VERSION = '0.08'; # VERSION
 
 use 5.010001;
 use strict;
 use warnings;
 
-our $DEBUG = 0;
-
-sub new {
-    my ($class, %attrs) = @_;
-    $attrs{default_section} //= 'GLOBAL';
-    $attrs{allow_bang_only} //= 1;
-    $attrs{enable_encoding} //= 1;
-    $attrs{enable_quoting}  //= 1;
-    $attrs{enable_bracket}  //= 1;
-    $attrs{enable_brace}    //= 1;
-    $attrs{enable_expr}     //= 0;
-    bless \%attrs, $class;
-}
-
-# borrowed from Parse::CommandLine. differences: returns arrayref. return undef
-# on error (instead of dying).
-sub __parse_command_line {
-    my $str = shift;
-
-    $str =~ s/\A\s+//ms;
-    $str =~ s/\s+\z//ms;
-
-    my @argv;
-    my $buf;
-    my $escaped;
-    my $double_quoted;
-    my $single_quoted;
-
-    for my $char (split //, $str) {
-        if ($escaped) {
-            $buf .= $char;
-            $escaped = undef;
-            next;
-        }
-
-        if ($char eq '\\') {
-            if ($single_quoted) {
-                $buf .= $char;
-            }
-            else {
-                $escaped = 1;
-            }
-            next;
-        }
-
-        if ($char =~ /\s/) {
-            if ($single_quoted || $double_quoted) {
-                $buf .= $char;
-            }
-            else {
-                push @argv, $buf if defined $buf;
-                undef $buf;
-            }
-            next;
-        }
-
-        if ($char eq '"') {
-            if ($single_quoted) {
-                $buf .= $char;
-                next;
-            }
-            $double_quoted = !$double_quoted;
-            next;
-        }
-
-        if ($char eq "'") {
-            if ($double_quoted) {
-                $buf .= $char;
-                next;
-            }
-            $single_quoted = !$single_quoted;
-            next;
-        }
-
-        $buf .= $char;
-    }
-    push @argv, $buf if defined $buf;
-
-    if ($escaped || $single_quoted || $double_quoted) {
-        return undef;
-    }
-
-    \@argv;
-}
-
-sub __read_file {
-    my $filename = shift;
-    open my $fh, "<", $filename
-        or die "Can't open file '$filename': $!";
-    binmode($fh, ":utf8");
-    local $/;
-    return ~~<$fh>;
-}
-
-sub __decode_json {
-    state $json = do {
-        require JSON;
-        JSON->new->allow_nonref;
-    };
-    my $res;
-    eval { $res = $json->decode(shift) };
-    if ($@) {
-        return [500, "Invalid JSON: $@"];
-    } else {
-        return [200, "OK", $res];
-    }
-}
-
-sub __decode_hex {
-    pack("H*", shift);
-}
-
-sub __decode_base64 {
-    require MIME::Base64;
-    MIME::Base64::decode_base64(shift);
-}
-
-sub _decode_expr {
-    my ($self, $val) = @_;
-    require Config::IOD::Reader::Expr;
-    no strict 'refs';
-    local *{"Config::IOD::Reader::Expr::val"} = sub {
-        my $arg = shift;
-        if ($arg =~ /(.+)\.(.+)/) {
-            return $self->{_res}{$1}{$2};
-        } else {
-            return $self->{_res}{ $self->{_cur_section} }{$arg};
-        }
-    };
-    my $res = Config::IOD::Reader::Expr::_parse_expr($val);
-    $self->_err("Can't decode expr: $res->[1]") if $res->[0] != 200;
-    $res->[2];
-}
-
-sub _err {
-    my ($self, $msg) = @_;
-    die join(
-        "",
-        @{ $self->{_include_stack} } ? "$self->{_include_stack}[0] " : "",
-        "line $self->{_linum}: ",
-        $msg
-    );
-}
-
-sub _push_include_stack {
-    require Cwd;
-
-    my ($self, $path) = @_;
-
-    # included file's path is based on the main (topmost) file
-    if (@{ $self->{_include_stack} }) {
-        require File::Spec;
-        my (undef, $dir, $file) =
-            File::Spec->splitpath($self->{_include_stack}[-1]);
-        $path = File::Spec->rel2abs($path, $dir);
-    }
-
-    my $abs_path = Cwd::abs_path($path) or return [400, "Invalid path name"];
-    return [409, "Recursive", $abs_path]
-        if grep { $_ eq $abs_path } @{ $self->{_include_stack} };
-    push @{ $self->{_include_stack} }, $abs_path;
-    return [200, "OK", $abs_path];
-}
-
-sub _pop_include_stack {
-    my $self = shift;
-
-    die "BUG: Overpopped _pop_include_stack" unless @{$self->{_include_stack}};
-    pop @{ $self->{_include_stack} };
-}
+use parent qw(Config::IOD::Base);
 
 sub _merge {
     my ($self, $section) = @_;
@@ -201,6 +32,17 @@ sub _merge {
     }
 }
 
+sub _init_read {
+    my $self = shift;
+
+    $self->SUPER::_init_read;
+    $self->{_res} = {};
+    $self->{_merge} = undef;
+    $self->{_num_seen_section_lines} = 0;
+    $self->{_cur_section} = $self->{default_section};
+    $self->{_arrayified} = {};
+}
+
 sub _read_string {
     my ($self, $str) = @_;
 
@@ -212,13 +54,11 @@ sub _read_string {
         qr/^;\s*!\s*(\w+)\s*/;
 
     my @lines = split /^/, $str;
-    $self->{_linum} = 0;
+    local $self->{_linum} = 0;
   LINE:
     for my $line (@lines) {
         $self->{_linum}++;
 
-        $self->{_last} = '';
-
         # blank line
         if ($line !~ /\S/) {
             next LINE;
@@ -239,7 +79,7 @@ sub _read_string {
                     if grep { $_ eq $directive }
                         @{$self->{disallow_directives}};
             }
-            my $args = __parse_command_line($line);
+            my $args = $self->_parse_command_line($line);
             if (!defined($args)) {
                 $self->_err("Invalid arguments syntax '$line'");
             }
@@ -257,7 +97,7 @@ sub _read_string {
                     $self->_err("Can't include '$path': $res->[1]");
                 }
                 $path = $res->[2];
-                $self->_read_string(__read_file($path));
+                $self->_read_string($self->_read_file($path));
                 $self->_pop_include_stack;
             } elsif ($directive eq 'merge') {
                 $self->{_merge} = @$args ? $args : undef;
@@ -275,7 +115,6 @@ sub _read_string {
 
         # section line
         if ($line =~ /^\s*\[\s*(.+?)\s*\](?: \s*[;#].*)?/) {
-            $self->{_last} = 'section';
             my $prev_section = $self->{_cur_section};
             $self->{_cur_section} = $cur_section = $1;
             $res->{$cur_section} //= {};
@@ -301,7 +140,7 @@ sub _read_string {
                 $val = $2;
             } elsif ($self->{enable_quoting} && $val =~ /^"/) {
                 $val =~ s/("[^"]*")\s*[;#].*/$1/; # strip comment (not perfect)
-                my $res = __decode_json($val);
+                my $res = $self->_decode_json($val);
                 if ($res->[0] != 200) {
                     $self->_err("Invalid JSON string");
                 }
@@ -309,7 +148,7 @@ sub _read_string {
                 $impenc++;
             } elsif ($self->{enable_bracket} && $val =~ /^\[/) {
                 $val =~ s/(\[[^\]]*\])\s*[;#].*/$1/; # strip comment (not perfect)
-                my $res = __decode_json($val);
+                my $res = $self->_decode_json($val);
                 if ($res->[0] != 200) {
                     $self->_err("Invalid JSON array");
                 }
@@ -317,7 +156,7 @@ sub _read_string {
                 $impenc++;
             } elsif ($self->{enable_brace} && $val =~ /^\{/) {
                 $val =~ s/(\{[^\]]*\})\s*[;#].*/$1/; # strip comment (not perfect)
-                my $res = __decode_json($val);
+                my $res = $self->_decode_json($val);
                 if ($res->[0] != 200) {
                     $self->_err("Invalid JSON object (hash)");
                 }
@@ -341,17 +180,17 @@ sub _read_string {
                         if grep { $_ eq $enc } @{$self->{disallow_encodings}};
                 }
                 if ($enc eq 'json') {
-                    my $res = __decode_json($val);
+                    my $res = $self->_decode_json($val);
                     if ($res->[0] != 200) {
                         $self->_err("Invalid JSON");
                     }
                     $val = $res->[2];
                 } elsif ($enc eq 'hex') {
                     $val =~ s/\s*[;#].*\z//; # shave comment
-                    $val = __decode_hex($val);
+                    $val = $self->_decode_hex($val);
                 } elsif ($enc eq 'base64') {
                     $val =~ s/\s*[;#].*\z//; # shave comment
-                    $val = __decode_base64($val);
+                    $val = $self->_decode_base64($val);
                 } elsif ($enc eq 'expr') {
                     $self->_err("Expr is not allowed (enable_expr=0)")
                         unless $self->{enable_expr};
@@ -389,32 +228,6 @@ sub _read_string {
     $res;
 }
 
-sub _init_read {
-    my $self = shift;
-    $self->{_res} = {};
-    $self->{_merge} = undef;
-    $self->{_include_stack} = [];
-    $self->{_num_seen_section_lines} = 0;
-    $self->{_cur_section} = $self->{default_section};
-}
-
-sub read_file {
-    my ($self, $filename) = @_;
-    $self->_init_read;
-    my $res = $self->_push_include_stack($filename);
-    die "Can't read '$filename': $res->[1]" unless $res->[0] == 200;
-    $res =
-        $self->_read_string(__read_file($filename));
-    $self->_pop_include_stack;
-    $res;
-}
-
-sub read_string {
-    my ($self, $str) = @_;
-    $self->_init_read;
-    $self->_read_string($str);
-}
-
 1;
 # ABSTRACT: Read IOD configuration files
 
@@ -430,7 +243,7 @@ Config::IOD::Reader - Read IOD configuration files
 
 =head1 VERSION
 
-This document describes version 0.07 of Config::IOD::Reader (from Perl distribution Config-IOD-Reader), released on 2015-01-03.
+This document describes version 0.08 of Config::IOD::Reader (from Perl distribution Config-IOD-Reader), released on 2015-03-12.
 
 =head1 SYNOPSIS
 
@@ -459,8 +272,7 @@ optimized for low startup overhead.
 
 =head1 EXPRESSION
 
-Config::IOD::Reader 0.05 adds experimental support for expression. This allows
-you to do something like this:
+Expression allows you to do things like:
 
  [section1]
  foo=1
@@ -478,7 +290,8 @@ And the result will be:
      section2 => {baz=>2, qux=>"greasemonkey", quux=>"greasemonkey 2"},
  }
 
-For safety, you'll need to set C<enable_expr> to 1 first.
+For safety, you'll need to set C<enable_expr> attribute to 1 first to enable
+this feature.
 
 The syntax of the expression (the C<expr> encoding) is not officially specified
 yet in the L<IOD> specification. It will probably be Expr (see
@@ -611,7 +424,11 @@ Read IOD configuration from a string. Die on errors.
 
 =head1 SEE ALSO
 
-L<IOD>, L<Config::IOD>, L<IOD::Examples>
+L<IOD> - specification
+
+L<Config::IOD> - round-trip parser for reading as well as writing IOD documents
+
+L<IOD::Examples> - sample documents
 
 =head1 HOMEPAGE
 
@@ -619,7 +436,7 @@ Please visit the project's homepage at L<https://metacpan.org/release/Config-IOD
 
 =head1 SOURCE
 
-Source repository is at L<https://github.com/sharyanto/perl-Config-IOD-Reader>.
+Source repository is at L<https://github.com/perlancar/perl-Config-IOD-Reader>.
 
 =head1 BUGS
 
